@@ -61,7 +61,13 @@ class RedshiftDDL(DDL):
 class Redshift(Database):
     ddl_validation_class = RedshiftDDL
 
-    sql_features = ["DROP CASCADE", "NO SET SCHEMA"]
+    def feature(self, feature):
+        return feature in (
+            "NEEDS CASCADE",
+            "CAN REPLACE VIEW",
+            "CANNOT CHANGE SCHEMA",
+            "CANNOT SPECIFY DDL IN SELECT",
+        )
 
     def create_engine(self, settings):
         # Create engine using the connect_args argument to create_engine
@@ -138,8 +144,7 @@ class Redshift(Database):
     def _create_table_select(
         self, table, schema, select, view=False, ddl=dict(), execute=True
     ):
-        """Returns SQL code for a create table from a select statement
-        """
+        """Returns SQL code for a create table from a select statement"""
         table = f"{schema+'.' if schema else ''}{table}"
         table_or_view = "VIEW" if view else "TABLE"
 
@@ -156,8 +161,7 @@ class Redshift(Database):
         return q
 
     def _create_table_ddl(self, table, schema, ddl, execute=False):
-        """Returns SQL code for a create table from a select statement
-        """
+        """Returns SQL code for a create table from a select statement"""
         if len(ddl["columns"]) == 0:
             raise DBError(
                 self.name, self.db_type, "DDL is missing columns specification"
